package PureMessage::Manager::Response;
use strict;
use warnings;
use HTML::Template;
use Carp;
use PureMessage::Manager::Config qw(%config %web_config);
use vars qw( %config %web_config);
use PureMessage::Manager::Widget;
use PureMessage::Manager::Widget::Error;
use PureMessage::Manager::Widget::TaskBox;
use PureMessage::Manager::Widget::Dialog;
use PureMessage::Manager::Lang;
use PureMessage::_sys;
use Data::Dumper;
use PureMessage::Manager::Utils qw( html_escape);
use vars qw($AUTOLOAD);
use HTML::Entities;
use Digest::SHA qw(sha256_hex);
# create widget factory
my $WIDGET = new PureMessage::Manager::Widget;
my $HOSTNAME;

sub AUTOLOAD
{
    my $self = shift;
    my ($method) = $AUTOLOAD =~ /.*::(.*)$/;
    if ($method ne 'DESTROY' ) {
        my $add = ( $method =~ s/^(add)// ) ? 1 : 0;
        eval "require PureMessage::Manager::Widget::$method";
        unless ( $@ ) {
            my %opts = @_;
            $opts{REQUEST} = $self->{REQUEST};
            my $w = $WIDGET->widget(CLASS =>"PureMessage::Manager::Widget::$method", %opts);
            if ($add) {
                $self->add_part($w, $opts{name} || $method);
            }
            $w->read_input($self->{REQUEST});
            return $w;
        }
        else {
            croak("Error loading $method: $@");
        }
    }
}

our $STANDARD_SAFE_ERROR = <<EOF;
Unable to initialize a library required to display this page. The most likely cause is a
misconfiguration of your database settings. Please check the manager error logs for
further details.
EOF


our %ERROR = ();

=item ERROR(["error_str"]) returns []

Error storage class for Request objects

=cut

sub ERROR {
  my $class = shift;
  
  if (scalar(@_) >= 1) {
      
    $ERROR{$$} = [] unless exists($ERROR{$$});
    push(@{$ERROR{$$}},@_);
  }
  return $ERROR{$$} if exists($ERROR{$$});
  return;
}

sub clear {
  delete($ERROR{$$});
}



#------------------------------------------------------------------------
# XXX
# This class should be made much more generic by moving much of this
# into a theme package that inherits this one.
#
# It would also be nice to be able to have more than one handler
# associated with an 'event'.
#-----------------------------------------------------------------------

my %template_options = (
    die_on_bad_params => 0,
    loop_context_vars => 1,
    );


#Insert a hidden input tag which contains CSRF token
my $salt = 'yRhx*7N3qndwH98g';
my $csrf_string = qq{$ENV{'SESSION_ID'}.$ENV{'REMOTE_USER'}.$salt};
my $token = sha256_hex($csrf_string);
my $tag = qq{<input name='CSRFtoken' type='hidden' value="$token">};

my $http_headers_sent = 0;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    
    # get the default page settings
    my %self = (@_, 'body_attrib' => _body_attributes());
    push(@{ $self{'html_head'} }, _default_head());
    
    # set the document content type and send the header unless told
    # not to.
    $self{'no_nav'} = 0 unless $self{'no_nav'};
    $self{'page_width'} = 600 unless $self{'page_width'};
    $self{'content-type'} = "text/html;charset=UTF-8" unless $self{'content-type'};
    $self{'html_meta_charset'} = 'iso-8859-1' unless $self{'html_meta_charset'};
    _print_header('content-type'=>$self{'content-type'}, @_)
        unless ($self{'delay-headers'} || $http_headers_sent);

    # redirect STDERR to STDOUT
    if ($web_config{'stderr_to_stdout'} || $self{'stderr_to_stdout'}) {
        open STDERR, ">&STDOUT";
        select STDOUT; $|=1;        
    }
    
    # keep a per-object copy of the template options
    my %to = %template_options;
    $self{'template_options'} = \%to;
    
    my $self = bless( \%self, $class);
    $self->communicate_safe_errors();
    return $self;
}

sub communicate_safe_errors {
    my $self = shift;
    my $errors = $self->ERROR();
    return unless $errors and ref($errors) eq 'ARRAY' and scalar(@$errors);
    foreach my $e (@$errors) {
      $self->raise_error($e);
    }
    $self->send();
    exit; # we had errors we cannot continue
}

sub raise_error {
    my $self = shift;
    my @data = @_;
    my $t = text(@data);
    if ($t ne $data[0]) {
        push(@{$self->{error}},$t); 
    } else {
        push(@{$self->{error}},join(' ',@data));
    }
    return 0;
}

sub get_system_hostname {
    my $self = shift;
    unless ($HOSTNAME) {
        $HOSTNAME = PureMessage::_sys::gethostname()
    }
    return $HOSTNAME;
}

sub error {
        my $self = shift;
        return unless $self->{error};
        return @{$self->{error}};
}

sub title {
    my $self = shift;
    my $title = shift;
    require PureMessage::Manager::Modules;
    my $m = PureMessage::Manager::Modules->new()->get_modules();
    my $module = $m->{$self->module()}->{title};
    $module .= " : " if $module;
    $self->{TITLE} = $module . text($title);
}

sub add_head {
    my $self = shift;
    my $data = shift;
    push(@{ $self->{'html_head'} }, $data);
}

# example:
#  $obj->add_template(
#      'file' => $file,
#      'data' => $data,
#      'name' => 'tmpl_name', <= this is optional
#  );
#
# return a reference to the template object
sub add_template {
    my $self = shift;
    my %p = @_;
    #push(@{ $self->{'templates'} }, \%p);

    #my $t = shift;
    my $tmp = HTML::Template->new(
        'filename' => $p{'file'},
        %{$self->{'template_options'}},
    );
    $tmp->param($p{'data'}) if $p{'data'};
    #print $tmp->output;
    my %part = (
                'object' => $tmp,
                'name'   => $p{'name'} || '',
               );
    
    push(@{ $self->{'parts'} }, \%part);
    return $tmp;
}

# return a reference to a named template object if
# direct access is needed.
sub get_template {
    my $self = shift;
    my $tname = shift;
    return $self->get_part($tname);
}

# rows are an arrayref with the following format:
# { 'type' => 'href' | 'custom',
#   'data' => <type-dependant>,
# }
#
# data format for href:
# ['description', 'link', 'img']
# image_file is optional
sub add_menu {
    my $self = shift;
    my $name = shift;
    my $rows = shift;
    push(@{ $self->{'menus'} },
        {'name' => $name, 'rows' => $rows});
}

sub add_taskmenu {
    my $self = shift;
    my $task = shift;
    my $pos = shift;
    if ( defined $pos ) {
        splice(@{ $self->{'taskmenus'}}, $pos, 0, {object => $task });
    }
    else {
        push(@{ $self->{'taskmenus'}}, {object => $task });
    }
}

sub add_help {
    my $self = shift;
    my @data = @_;
    push(@{ $self->{'helpmenu'}}, \@data);
}

# Adds a menu that is based on a template
# The template takes care of producing rows if needed so this just
# creates a single row menu.
sub add_template_menu {
    my $self = shift;
    my $name = shift;
    my %ops = @_;
    my $t = HTML::Template->new(
                                'filename' => $ops{'file'},
                                %{$self->{'template_options'}},
                               );
    $t->param($ops{'data'}) if $ops{'data'};
    $t->param(CSRFtoken => $token);
    my $data = $t->output;

    # make this look like a regular custom task menu
    my $mdata = {
    'title' => $name,
    'buttons' => [[{'type' => 'custom', 'text' => $data}]],
    };
    $self->add_taskmenu($mdata);
}

sub add_part {
    my $self = shift;
    my $part = shift;
    my $name = shift || '';
    push(@{ $self->{'parts'} }, {'object' => $part, 'name' => $name});
    return $part;
}

sub insert_part {
    my $self = shift;
    my $part = shift;
    my $name = shift || '';
    my $pos = shift || 0;
    splice ( @{$self->{'parts'}}, $pos, 0, {'object' => $part, 'name' => $name});
    return $part;
}

sub unshift_part {
    my $self = shift;
    my $part = shift;
    my $name = shift || '';
    unshift(@{ $self->{'parts'} }, {'object' => $part, 'name' => $name});
    return $part;
}


sub dialog {
    my $self = shift;
    my $dialog = new PureMessage::Manager::Widget::Dialog(@_);
    $self->add_part($dialog,'dialog');
}

sub error_dlg {
    my $self = shift;
    my @errors = $self->error();
    return unless @errors;
    my $e = new PureMessage::Manager::Widget::Error(error => \@errors, css_class => 'normal');
    my $dialog = new PureMessage::Manager::Widget::Dialog( style => 'dlg_error',
                                                           msg => $e->output(),
                                                           name => 'dialog',
                                                           title => 'An Error Occurred.',
                                                         );
    $self->add_part($dialog,'dialog');
    $self->{error} = undef; # clear this since we have already gotten them for display
}

# TODO don't duplicate this!!
sub error_dlg_str {
    my $self = shift;
    my @errors = $self->error();
    return unless @errors;
    my $e = new PureMessage::Manager::Widget::Error(error => \@errors, css_class => 'normal');
    my $dialog = new PureMessage::Manager::Widget::Dialog( style => 'dlg_error',
                                                           msg => $e->output(),
                                                           name => 'dialog',
                                                           title => 'An Error Occurred.',
                                                         );
    $self->{error} = undef; # clear this since we have already gotten them for display
    return $dialog->output();
}

sub notification_str {
    my $self = shift;
    return "<b>" . text(shift(@_)) . "</b>";
}

# return a reference to a named part object if
# direct access is needed.
sub get_part {
    my $self = shift;
    my $pname = shift;
    foreach my $part (@{ $self->{'parts'} }) {
        return $part->{'object'} if $part->{'name'} eq $pname;
    }
}

sub print {
    my $self = shift;
    my @data = @_;
    if(@data == 1) {
        return $self->add_part($data[0]);
    } else {
        return(map { $self->add_part($_) } @data);
    }
}

sub print_direct {
    my $self = shift;
    print @_;
}

# Note: The table style used in this function is spread all over the place
# it would be nice to have a common (flexible) interface to this to keep
# maintenance simple.
sub print_notification {
    my $self = shift;
    my $start = <<'EOT';
<table cellpadding=4 cellspacing=0 width="100%">
<tr>
<td width="1" class="infoBox">
    <img src="/img/info.gif">
</td>
<td class="infoBox">        
EOT
    my $end = <<'EOT';
</td></tr>
</table><br>
EOT
    my $msg = shift || '';
    $self->print($start . $msg . $end);
}

sub body_attrib {
    my $self = shift;
    if (@_ < 2 ) {
        return $self->{'body_attrib'}{$_[0]};
    }
    else {
        $self->{'body_attrib'}{$_[0]} = $_[1];
    }
}

sub set_handler {
    my $self = shift;
    my $hname = shift;
    my $handler = shift;
    $self->{'handler'}{$hname} = $handler;
}

sub new_template {
    my $self = shift;
    my $tmpl = shift;
    my $head = HTML::Template->new(
            'filename' => $tmpl,
            path => $web_config{tmpl_path},
            %{$self->{'template_options'}},
    );
    return $head;
}

# put it all together and send the response
sub send {
    my $self = shift;

    # prevent recurive calls
    croak "Send was already called." if $self->{'in_send'};
    $self->{'in_send'} = 1;
    
    eval{
        
        # http header (can be overridden)
        if ($self->{'delay-headers'}) {
            unless ($self->_do_handler('http_headers')) {
                _print_header(%$self);
            }
        }
       
        # html head section
        # (can be overridden)
        unless ($self->_do_handler('html_head')) {
            my $head = $self->new_template("_htmlheader.html");
        $head->param(HTML_META_CHARSET => $self->{'html_meta_charset'});
            my @extra = map { { CONTENT => $_ } } @{ $self->{'html_head'} };
            if (my $refresh = $self->{REQUEST}->query('refresh_int')) {
                $refresh = encode_entities($refresh,'&<>"\/\'');
                my $url = $self->{REQUEST}->query('refresh_url') || $ENV{REQUEST_URI};
                $url .= "?" unless $url =~ /\?/;
                $url .= "&refresh_int=$refresh" unless $url =~ /refresh_int/;
                $url .= "&refresh_url=" . $self->{REQUEST}->query('refresh_url') 
                    if $self->{REQUEST}->query('refresh_url') and $url !~ /refresh_url/;
                push @extra, { CONTENT => "<META HTTP-EQUIV=Refresh CONTENT=\"$refresh; URL=$url\">" };
            }
            push @extra, {CONTENT => <<EOT};
<STYLE>
td.leftMenu {
    background-color: #fff;
}
</STYLE>
EOT
            $head->param(HEAD => \@extra);
            print $head->output();
        }
        
        
        unless ($self->_do_handler('content_start') or $self->{no_wrap}) {
            $self->_init_sidebars();
            
            my $start = $self->new_template("_header.html");
            $start->param(TITLE    => $self->{TITLE},
                          TOP_NAV  => $self->_primary_nav(),
                          MENU     => $self->_sidebars() || [],
                          BANNER   => text("banner"),
                          BODYATTR => $self->_has_handler('body_attrib') ? $self->_do_handler('body_attrib') : '',
                          HOSTNAME => $self->get_system_hostname(),
                          USER     => $ENV{'ANONYMOUS_USER'} ? "Anonymous user" : $ENV{'REMOTE_USER'},
                          AUTH     => $ENV{'SSL_USER'} ? "(SSL certified)" :
                                      $ENV{'LOCAL_USER'} ? "(Local user)" : "",
                        );
            print $start->output();
        }
    
        # display any errors 
        if ($self->error() ) {
                $self->_print_errors($self->error);
        }
        # send CGI response components
        foreach my $p (@{$self->{'parts'}}) {
            $self->_print_part($p);
        }
        
        $self->_do_handler('content_end');
        
        if($self->{debug}) {
                $self->_print_debug();
        }
        
        # and the page footer (can be overridden)
        unless ($self->_do_handler('body_end') or $self->{no_wrap}) {
            my $end = $self->new_template("_footer.html");
            print $end->output();
        }
    };
    if ($@) {
        print STDERR "An error occured while sending the response. $@";
    }
    $self->{'in_send'} = 0;
}

sub _print_debug {
    my $self = shift;
    print $self->_debug_hash('CGI Input', $self->{REQUEST}{query});
    print $self->_debug_hash('Response Object ' . ref($self),$self);
    print $self->_debug_hash('ENV Hash', $self->{REQUEST}{env});
    print $self->_debug_hash('Web Config',\%web_config);
    print $self->_debug_hash('Config',\%config);
    $self->_debug_part();
}

sub _debug_hash {
    my $this = shift;
    my ($title,$hash) = @_;
    return dump_hash($title,$hash);
}

sub _print_errors {
    my $self = shift;
    my @errors = @_;
  
    my $widget = new PureMessage::Manager::Widget::Error(error => \@errors);
    print $widget->output();
}


sub _sidebars {
    my $self = shift;
    return if $self->{'no_sidebar'};
    my @bars;
    foreach my $menu (@{ $self->{'menus'} }) {
        my $content = "";
        foreach my $row (@{ $menu->{'rows'}}) {
            if (lc($row->{'type'}) eq 'href') {
                $content .= _format_href(@{ $row->{'data'} }) . "<br/>";
            }
            elsif (lc($row->{'type'}) eq 'custom') {
                $content .= ${ $row->{'data'} } . "<br/>";
            }
        }
        push @bars, { 'title' => $menu->{'name'}, 'content' => $content };
    }
    foreach my $p (@{ $self->{'taskmenus'} }) {
        if ($p && $p->{'object'}) {
            if (ref($p->{'object'}) eq 'HASH') {
                my $content = "";
                foreach (@{$p->{'object'}{'buttons'}[0]}) {
                    my $tg = $_->{target} 
                             ? "target='$_->{target}'"
                             : "";
                    $content .= "<a $tg href='$_->{href}'>$_->{text}</a><br />\n" if $_->{href};
                    $content .= $_->{text} if $_->{type} eq "custom";
                }
                push @bars, { title => $p->{'object'}{'title'}, 
                              content => $content,
                              toolbar => $p->{'object'}{'tb'},
                            };
            }
            elsif (ref($p->{'object'}) && $p->{'object'}->can('output')) {
                my $title = $p->{'object'}->can('title') ? $p->{'object'}->title() : "";
                push @bars, { title => $title, 
                              content => $p->{'object'}->output(),
                              toolbar => $p->{'object'}{'tb'}, 
                            };
            }
            elsif (ref(\($p->{'object'})) eq 'SCALAR') {
                push @bars, { title => '', 
                              content => $p->{'object'},
                              toolbar => $p->{'object'}{'tb'},
                            };
            } 
        }
    }
    if ($self->{'helpmenu'}){
        my $content;
        foreach my $p (@{ $self->{'helpmenu'} }) {
            $content .= _format_help(@$p) . "<br/>";
        }
        push @bars, { title => "Help", 
                      content => $content,
                    };
    }
    # attempt to merge any menus with the same title
    my @bars_final;
    my %titles;
    foreach my $b (@bars) {
        if (exists($titles{$b->{title}})) {
          $bars_final[$titles{$b->{title}}]->{content} .= $b->{content};
        } else {
          push(@bars_final,$b);
          $titles{$b->{title}} = $#bars_final;
        }
    }
    return \@bars_final;
}

sub _body_attributes {
    my %a = (
        'class' => 'body',
    );
    return \%a;
}

sub _default_head {
    my $user = $ENV{USER};
    my $hostname = get_system_hostname();
    my $status = $user ? "$user logged into PureMessage Manager on $hostname" : "";
    return <<"EOT";
<script LANGUAGE=\"JavaScript\">
defaultStatus=\"$status\";
</script> 
EOT
}

sub _print_header {
    my %options = @_;

    # workaround a bug where IE causes "file not found" errors when a content-disposition
    # header is sent over SSL. XXX remove this when IE is fixed.
    unless (grep { lc($_) eq 'content-disposition' } keys %{$options{'http-headers'}} ) {
      print "Pragma: no-cache\n";
      print "Expires: -1\n";
      print "Cache-Control: no-cache\n";
    }
    print "Content-Type: $options{'content-type'}\n";
    foreach my $header (keys %{$options{'http-headers'}}) {
    print "$header: " . $options{'http-headers'}->{$header} . "\n";
    }
    print "\n";
    $http_headers_sent = 1;
}

sub _primary_nav {
    my $self = shift;
    unless ( $self->{'no_nav'} ) {
        require PureMessage::Manager::Modules;
        require PureMessage::Manager::Users;
        PureMessage::Manager::Users::open_db();
        my $current = $self->module();
        my $m = PureMessage::Manager::Modules->new();
        my $mods = $m->get_modules;
        my @buttons;
        foreach ( $m->get_module_names() ) {
            next unless $mods->{$_}{title};
            next unless PureMessage::Manager::Users::can_user_access($ENV{REMOTE_USER}, $_);
            push @buttons, {  URL => "/$_/", LABEL => $mods->{$_}{title}, 
                              ICON => $mods->{$_}{icon}, DESC => $mods->{$_}{description},
                              SEL => ( "$_" eq $current ) ? "Sel" : "",
                            };
        }
        return \@buttons;
    }
    return [];
}
# sidebars: pos means the menu will be inserted into the sidebar at that absolute index
# so this one will be inserted as the 3'rd one from 0. Items in that menu are sorted 
# lexically by name so "First" will first in the menu and "Second" second.
#
#<module Quarantine>
#        pos=3
#        <menu Enduser>
#            <item g_enduser_z_black_list>
#                title = "Second"
#                url = "/Quarantine/Enduser/list.cgi?l=1&n=enduser-black-list"
#            </item>
#            <item g_enduser_a_black_list>
#                title = "First"
#                url = "/Quarantine/Enduser/list.cgi?l=1&n=enduser-black-list"
#            </item>
#        </menu>
#    </module>
#
#<module test>
#  <menu>
#     path="/Services/Manager"
#     ...
#  </menu>
#</module>
#
# Will only show on pages that match the listed path
# 
sub _init_sidebars {
    my $self = shift;
    unless ( $self->{'no_sidebar'} ) {
        require PureMessage::Manager::Modules;
        my $current = $self->module();
        my $m = PureMessage::Manager::Modules->new();
        my $i = 0;
        if ( my $menus = $m->get_menus($current) ) {
            while ( my ( $menu, $data ) = each %{$menus} ) {
                if(defined($data->{path})) {
                  # then only for this path
                  my @paths = ref($data->{path}) eq 'ARRAY' ? @{$data->{path}} : ($data->{path});
                  next unless grep { $ENV{SCRIPT_NAME} =~ /^\Q$_\E/ } @paths;
                }
                my @menu;
                foreach ( sort keys %{$data->{item}} ) {
                    push @menu, { type=>'text', text=>$data->{item}{$_}{title}, href=>$data->{item}{$_}{url} };
                }
                my $menu_title = exists($data->{title}) ? $data->{title} : $menu;
                $self->add_taskmenu({ title => $menu_title, buttons => [ \@menu ], }, $data->{pos} ? $data->{pos} : $i++);
            }
        }
        my $sidebars = $m->get_sidebars($current);
        foreach my $path ( keys %{$sidebars} ) {
            my @path = ($current, split (/\:\:/, $path) );
            my $module = join("::", @path[0..$#path-1]);
            my $call = join("::", @path);
            eval <<EOT;
            require PureMessage::Manager::Response::$module;
            PureMessage::Manager::Response::$call(\$self);
EOT
        }
    }
}

sub _print_template{
    my $self = shift;
    my $t = shift;
    my $tmp = HTML::Template->new(
        'filename' => $t->{'file'},
        %{$self->{'template_options'}},
    );
    $tmp->param($t->{'data'});
    print $tmp->output;
}

sub _debug_part{
    my $self = shift;
    
    foreach my $p (@{$self->{'parts'}}) {
    
      # output debug info for each part if we can
      if ($p && $p->{'object'}) {
          if (ref($p->{'object'}) && $p->{'object'}->can('output_debug')) {
              print $p->{'object'}->output_debug();
          }
      }

    }
}

sub _print_part{
    my $self = shift;
    my $p = shift;
    my $output;
    if ($p && $p->{'object'}) {
        if (ref($p->{'object'}) && $p->{'object'}->can('output')) {
            # propogate our debug setting to this object if it will allow it
            $p->{'object'}->debug($self->{debug}) if $p->{'object'}->can('debug');
            my $output = $p->{'object'}->output();
            $output =~ s!(<form\b[^>]*>)!$1$tag!ig;
            print $output;
        }
        elsif (ref(\($p->{'object'})) eq 'SCALAR') {
            $output = $p->{'object'};
            $output =~ s!(<form\b[^>]*>)!$1$tag!ig;
            print $output;
        }
    }
}

sub _do_handler {
    my $self = shift;
    my $handler = shift;
    if ($self->{'handler'}{$handler} ) {
        $self->{'handler'}{$handler}();
        return 1;
    }
    return 0;
}

sub _has_handler { shift->{handler}{shift} }

sub _format_href {
    my @data = @_;
    my $img = '';
    if ($_[2]) {
        $img = '<img border="0" align="middle" style="margin-bottom: 3px" src="' . $_[2] . '">';
    }
    my $line= "<a class=\"sidebar\" href=\"$data[1]\">$img $data[0]</a>\n";
    return $line;
}

# called indirectly from add_help() as:
# add_help(page => "foo", tab => "bar", text => "Click Me")
# 
sub _format_help {
    my %params = @_;

    $params{text} = "Help" unless $params{text};

    my $img = '<img border="0" width="15" height="14" src="/img/xlink.gif">';
    my $open = "window.open(this.href, 'docs', 'width=850, height=600, toolbar=yes, resizable=yes'); return false;";
    my $href = "/Help/help.html?tab=$params{tab}&page=$params{page}";
    my $id = "$params{tab}_$params{page}";
    my $line= <<EOT;
<a class="sidebar" id="$id" href="$href" onClick="$open">$params{text} $img</a>
EOT
    return $line;
}

# action, type, object, \%params, [module]
sub log {
    my $self = shift;
    my ($action, $type, $object, $params, $m) = @_;
    $object ||= "";
    return if (!$web_config{'log'});
    my $base_remote_user = $ENV{'BASE_REMOTE_USER'} 
                         ? $ENV{'BASE_REMOTE_USER'}
                         : $ENV{'REMOTE_USER'};
    my $remote_user = $ENV{'REMOTE_USER'};
                                    
    my $webmin_logfile = $web_config{'activity_log'};
    return unless $webmin_logfile;
    $m ||= $self->module();
    if ($web_config{'logclear'}) {
        # check if it is time to clear the log
        my @st = stat("$webmin_logfile.time");
        my $write_logtime = 0;
        if (@st) {
            if ($st[9] + $web_config{'logtime'} * 60 * 60 < time()) {
                # clear logfile and all diff files
                system("rm -f $ENV{'WEBMIN_VAR'}/diffs/* 2>/dev/null");
                unlink($webmin_logfile);
                $write_logtime = 1;
            }
        }
        else {
            $write_logtime = 1; 
        }
        if ($write_logtime) {
            open(LOGTIME, ">$webmin_logfile.time");
            print LOGTIME time(),"\n";
            close(LOGTIME);
        }
    }
    # should logging be done at all?
    my %logusers = map { $_ => 1 } split(/\s+/, $web_config{'logusers'} || "");
    my %logmods = map { $_ => 1 } split(/\s+/, $web_config{'logmods'} || "");
    return if $web_config{'logusers'} and not $logusers{$base_remote_user};
    return if $web_config{'logmodules'} and not $logmods{$m};

    # log the action
    my $now = time();
    my @tm = localtime($now);
    my $script_name = $0 =~ /([^\/]+)$/ ? $1 : '-';
    my $id = sprintf "%d.%d.%d", $now, $$, $self->{action_id_count} || 0;
    $self->{action_id_count}++;
    my $line = sprintf "%s [%2.2d/%s/%4.4d %2.2d:%2.2d:%2.2d] %s %s %s %s %s \"%s\" \"%s\" \"%s\"",
        $id, $tm[3], text("smonth_" . ($tm[4] + 1)), $tm[5] + 1900,
        $tm[2], $tm[1], $tm[0],
        $remote_user, $self->session_id() ? $self->session_id() : '-',
        $ENV{'REMOTE_HOST'},
        $m, $script_name,
        $action, $type ne '' ? $type : '-', $object ne '' ? $object : '-';

    foreach my $k (sort { $a cmp $b } keys %{$_[3]}) {
        my $v = $params->{$k};
        if ($v eq '') {
            $line .= " $k=''";
        }
        elsif (ref($v) eq 'ARRAY') {
            foreach my $vv (@$v) {
                next if (ref($vv));
                $vv =~ s/([\'\"\\\r\n\t\%])/sprintf("%%%2.2X",ord($1))/ge;
                $line .= " $k='$vv'";
            }
    }
        elsif (!ref($v)) {
            foreach my $vv (split(/\0/, $v)) {
                $vv =~ s/([\'\"\\\r\n\t\%])/sprintf("%%%2.2X",ord($1))/ge;
                $line .= " $k='$vv'";
            }
        }
    }
    open(WEBMINLOG, ">>$webmin_logfile");
    print WEBMINLOG $line,"\n";
    close(WEBMINLOG);
    
    if ($web_config{'logfiles'}) {
        # Find and record the changes made to any locked files
        my $i = 0;
        mkdir("$ENV{'WEBMIN_VAR'}/diffs", 0700);
        foreach my $d (@{$self->{diff_file}}) {
            open(DIFFLOG, ">$ENV{'WEBMIN_VAR'}/diffs/$id.$i");
            print DIFFLOG "$d->{'type'} $d->{'object'}\n";
            print DIFFLOG $d->{'data'};
            close(DIFFLOG);
            $i++;
        }
        $self->{diff_file} = [];
    }
}

# log_diff(type, object, data)
sub log_diff {
    my $self = shift;
    if ($web_config{'logfiles'}) {
        $self->{diff_file} = [] unless $self->{diff_file};
        push(@{$self->{diff_file}},
            { 'type' => $_[0], 'object' => $_[1], 'data' => $_[2] } );
    }
}

sub session_id {
    return $ENV{'SESSION_ID'};
}

sub module {
    my $self = shift;
    my $current = $ENV{SCRIPT_NAME};
    $current = $1 if $current=~/\/([^\/]*)/;
    return $current;
}

1;




